""" Static definitions for all the models.

All the models were generated using [sqlacodegen](https://github.com/agronholm/sqlacodegen).

To use a model `import` it and use it as follows:
```python
from easyfatt_db_connector.orm.static import TDocRighe

with engine.connect() as conn:
    query = TDocRighe.__table__.select()
    for row in conn.execute(query):
        print(row)
```
"""
from .models import *


from dataclasses import dataclass
from typing import Optional
import lxml.etree as ET

from easyfatt_db_connector.xml.common import Field, XMLMapper
from easyfatt_db_connector.xml.vat_code import VatCode


@dataclass(init=False, repr=False)
class Product(XMLMapper):
    """ Campi nelle righe del documento (elementi <Row>). """
    __xml_name__ = "Row"

    __xml_mapping__ = {
        "code": "Code",
        "supplier_code": "SupplierCode",
        "description": "Description",
        "quantity": "Qty",
        "unit_measure": "Um",
        "size": "Size",
        "color": "Color",
        "lot": "Lot",
        "expiry_date": "ExpiryDate",
        "serial": "Serial",
        "price": "Price",
        "discounts": "Discounts",
        "eco_fee": "EcoFee",
        "vat_info": Field(VatCode, tag="VatCode"),
        "total": "Total",
        "withholding_tax": "WithholdingTax",
        "stock": "Stock",
        "notes": "Notes",
        "commission_percentage": "CommissionPerc",
    }

    code: str = ""
    """ Codice prodotto.
    
    Example:
        ```xml
        <Row>
            <Code>...</Code>
        </Row>
        ```
    """

    supplier_code: str = ""
    """ Codice prodotto del fornitore.
    
    Example:
        ```xml
        <Row>
            <SupplierCode>...</SupplierCode>
        </Row>
        ```
    """

    description: str = ""
    """ Descrizione prodotto o nota.
    
    Example:
        ```xml
        <Row>
            <Description>...</Description>
        </Row>
        ```
    """

    # I know it's strange, but in the XML schema the default is 0: 
    # `<xs:element name="Qty" type="xs:decimal" minOccurs="0" default="0"/>`
    quantity: int = 0
    """ Quantità prodotti.
    
    Example:
        ```xml
        <Row>
            <Qty>...</Qty>
        </Row>
        ```
    """

    unit_measure: str = ""
    """ Unità di misura della quantità.
    
    Example:
        ```xml
        <Row>
            <Um>...</Um>
        </Row>
        ```
    """

    size: str = ""
    """ Taglia (usato nel settore dell'abbigliamento).
    
    Example:
        ```xml
        <Row>
            <Size>...</Size>
        </Row>
        ```
    """

    color: str = ""
    """ Colore (usato nel settore dell'abbigliamento).
    
    Example:
        ```xml
        <Row>
            <Color>...</Color>
        </Row>
        ```
    """

    lot: str = ""
    """ Lotto.
    
    Example:
        ```xml
        <Row>
            <Lot>...</Lot>
        </Row>
        ```
    """

    expiry_date: str = "2999-12-31"
    """ Data scadenza lotto [Data].
    
    Example:
        ```xml
        <Row>
            <ExpiryDate>...</ExpiryDate>
        </Row>
        ```
    """

    serial: str = ""
    """ Codice seriale
    
    Example:
        ```xml
        <Row>
            <Serial>...</Serial>
        </Row>
        ```
    """

    price: float = 0
    """ Prezzo unitario [Valuta].
    
    Example:
        ```xml
        <Row>
            <Price>...</Price>
        </Row>
        ```
    """

    discounts: str = ""
    """ Sconti (es: "20+5.5%").
    
    Example:
        ```xml
        <Row>
            <Discounts>...</Discounts>
        </Row>
        ```
    """

    eco_fee: float = 0
    """ Importo dell'eco-contributo unitario associato all'articolo.
    
    Example:
        ```xml
        <Row>
            <EcoFee>...</EcoFee>
        </Row>
        ```
    """

    vat_info: Optional["VatCode"] = None
    """ Informazioni sull'IVA applicata al prodotto.
    
    Example:
        ```xml
        <Row>
            <VatCode>...</VatCode>
        </Row>
        ```
    """

    total: str = ""
    """ Importo complessivo della riga [Valuta]. Ignorato in importazione.
    
    Example:
        ```xml
        <Row>
            <Total>...</Total>
        </Row>
        ```
    """

    withholding_tax: Optional[bool] = None
    """ Ritenuta d'acconto applicata.
    
    Example:
        ```xml
        <Row>
            <WithholdingTax>false</WithholdingTax>
        </Row>
        ```
    """

    stock: Optional[bool] = None
    """ Movimentazione magazzino.
    
    Example:
        ```xml
        <Row>
            <Stock>true</Stock>
        </Row>
        ```
    """

    notes: str = ""
    """ Note della riga.
    
    Example:
        ```xml
        <Row>
            <Notes>...</Notes>
        </Row>
        ```
    """

    commission_percentage: float = 0
    """ Percentuale provvigione agente.
    
    Example:
        ```xml
        <Row>
            <CommissionPerc>...</CommissionPerc>
        </Row>
        ```
    """


if __name__ == "__main__":
    xml = """
        <Row>
            <Code>0012</Code>
            <Description>Clip con catenella</Description>
            <Qty>10</Qty>
            <Um>pz</Um>
            <Price>4.62</Price>
            <Discounts/>
            <VatCode Perc="20">My20VAT</VatCode>
            <Total>46.2</Total>
            <Stock>false</Stock>
        </Row>
    """
     
    product = Product.from_xml_string(xml)
    print(f"Example of <Product> as object: {product}\n")
    print(f"Product description is        : {product.description}\n")
    print(f"Product VatCode is            : {product.vat_info}")
    print(f"Product VatCode percentage is : {product.vat_info.percentage}")
    print(f"Product VatCode codeID is     : {product.vat_info.code}")
    print(f"Product VatCode class is      : {product.vat_info.vat_class}\n")

    # TODO: fix this
    # print(f"Example of <Product> as XML   : {product.to_xml_string()}\n")


import sqlalchemy
import sqlalchemy_firebird

from sqlalchemy.orm.decl_api import DeclarativeMeta

__all__ = ["table_factory"]

Base = sqlalchemy.orm.declarative_base()

def table_factory(name: str, engine: sqlalchemy.Engine) -> DeclarativeMeta:
	""" Dynamically create a `DeclarativeMeta` class which instance can be used for SQL queries.
	
	This function can be useful in the following cases:
	- The static declaration is somehow broken for a specific table
	- A table is not present in the static declaration

	!!! Warning
		Always prefere static classes (`easyfatt_db_connector.orm.static`) instead of dynamically generated ones.
		This gives you intellisense over the table and allows for faster and easier development.

	Args:
		name (str): The name of the table
		engine (sqlalchemy.Engine): The engine object.

	Returns:
		sqlalchemy.orm.decl_api.DeclarativeMeta: The clas that will be used to perform queries on.
	
	!!! Example
		The following example shows how to execute queries on a hypotetical table `TTable` which 
		is not present among the static definitions:
		```
		TTable = table_factory("TTable", engine)
		with engine.connect() as conn:
			query = TTable.__table__.select()
			for row in conn.execute(query):
				print(row)
		```
	"""
	return type(name, (Base,),{
		"__table__": sqlalchemy.Table(name, Base.metadata, autoload_with=engine),
		"__tablename__": name
	})


from pathlib import Path
from typing import Optional, Union

from easyfatt_db_connector.xml.root import EasyfattXML

__all__ = ["read_xml", "EasyfattXML"]

def read_xml(filename: Optional[Union[str, Path]] = None, text: Optional[str] = None, convert_types=True) -> EasyfattXML:
    """ Legge un file XML e lo converte in un oggetto `EasyfattXML`. 
    
    Args:
        filename (Union[str, Path], optional): Percorso del file XML da leggere. Defaults to None.
        text (str, optional): Testo del file XML da leggere. Defaults to None.
        convert_types (bool, optional): Se `True`, converte i valori dei tag in tipi Python. Defaults to True.
    
    Raises:
        ValueError: Se non viene fornito né un filename né un testo, oppure se vengono forniti entrambi.

    Returns:
        EasyfattXML: Oggetto `EasyfattXML` contenente i dati del file XML.
    """
    if (filename is not None and text is not None) or (filename is None and text is None):
        raise ValueError("You must provide a filename or a text")

    xml_string = ""
    if filename is not None:
        with open(filename, "r", encoding='utf-8-sig') as f:
            xml_string = f.read().strip()

    elif text is not None:
        xml_string = text.strip()
    
    return EasyfattXML.from_xml_string(bytes(xml_string, encoding='utf-8'), convert_types=convert_types)



from __future__ import annotations

import sys
import logging
from typing import Literal

from pathlib import Path
import shutil

from urllib.parse import urlparse
import requests

from easyfatt_db_connector.constants import DEFAULT_FIREBIRD_LOCATION

RELEASE_URL = "https://api.github.com/repos/FirebirdSQL/firebird/releases"
PROCESSOR_ARCHITECTURE = 32 << bool(sys.maxsize >> 32)

logger = logging.getLogger(__name__)

__all__ = ["get_available_releases", "download"]

def get_available_releases() -> list[str]:
    """ Retrieves the available releases from the FirebirdSQL repository.

    Returns:
        releases (list[str]): List of available release tags.
    """
    response = requests.get(RELEASE_URL)

    return [
        release["tag_name"]
        for release in response.json()
        if list(
            filter(
                lambda a: "_x64_embed.zip" in a["name"]
                or "_Win32_embed.zip" in a["name"],
                release["assets"],
            )
        )
    ]


def download(
    tag_name: Literal["R2_5_8", "R2_5_9"] | None = None,
    firebird_path: Path = DEFAULT_FIREBIRD_LOCATION,
    architecture: Literal[32, 64] | None = None,
) -> Path:
    """Downloads a local copy for the Firebird Embedded DB.

    Args:
        tag_name ('R2_5_8' | 'R2_5_9', optional): The tag name of the release (see https://github.com/FirebirdSQL/firebird/releases).
        firebird_path (Path, optional): Path where the firebird database will be installed. Defaults to `~/.cache/firebird-embedded/`.
        architecture (Literal[32, 64] | None, optional): Desired processor architecture (32 or 64 bit, None for automatic discovery). Defaults to None.

    Raises:
        Exception: No available releases were found
        Exception: Release '{tag_name}' not found. Available releases are {available_releases}.
        Exception: Unexpected number of downloads found: {download_no}. Expected 1.

    Returns:
        firebird (Path): Path to the Firebase Embedded folder
    """
    available_releases = get_available_releases()

    desired_architecture = (
        architecture if architecture is not None else PROCESSOR_ARCHITECTURE
    )

    if len(available_releases) == 0:
        raise Exception("No available releases were found")
    elif tag_name is not None and tag_name not in available_releases:
        raise Exception(f"Release '{tag_name}' not found. Available releases are {', '.join(available_releases)}.")

    release_name = tag_name if tag_name is not None else available_releases[0]
    release = requests.get(
        f"https://api.github.com/repos/FirebirdSQL/firebird/releases/tags/{release_name}"
    ).json()
    assets = release["assets"]

    # Retrieve the download url for the desired architecture
    remote_package_urls = [
        asset["browser_download_url"]
        for asset in assets
        if ("_x64_embed.zip" if desired_architecture == 64 else "_Win32_embed.zip")
        in asset["name"]
    ]

    if len(remote_package_urls) != 1:
        raise Exception(
            f"Unexpected number of downloads found: {', '.join(remote_package_urls)}. Expected 1."
        )

    zip_file_path: Path = (
        firebird_path / str(urlparse(remote_package_urls[0]).path).split("/")[-1]
    )
    version_file = firebird_path / "version.txt"

    try:
        firebird_path.mkdir(parents=True)
    except FileExistsError:
        # Check if version is the same
        if version_file.exists() and version_file.read_text() == zip_file_path.stem:
            return firebird_path

        # Remove all old files
        for child in firebird_path.glob("*"):
            if child.is_dir():
                shutil.rmtree(child)
            else:
                child.unlink()

    # Download the archive and extract it
    response = requests.get(remote_package_urls[0], allow_redirects=True)
    with open(zip_file_path, "wb") as zip_file:
        zip_file.write(response.content)

    shutil.unpack_archive(zip_file_path, firebird_path)
    zip_file_path.unlink()

    # Write the version file
    version_file.write_text(zip_file_path.stem)

    return firebird_path


if __name__ == "__main__":
    print(download())
    # print(download("R2_5_8"))
    # print(download("R2_5_9", architecture=32))



# import os as _os
# import xml.dom.minidom as minidom
from typing import get_type_hints

import lxml.etree as ET

from easyfatt_db_connector.core.exceptions import TypeConversionError

from .fields import BaseField, Field, FieldGroup


class XMLMapper(object):
    """Base class for XML mappers.

    - Use the `__xml_mapping__` class attribute to map XML elements to class attributes.
    - Use the `__xml_name__` class attribute to override the name of the XML element the class refers to.
    """

    __xml_name__: str = ""
    """ Customize the name of the XML tag. 
    
    If not specified, it will be the set to the name of the class.
    """

    __xml_mapping__: dict[str, str] = None

    def __str__(self) -> str:
        attributes = [
            (f"{attr}='{value}'" if type(value) == str else f"{attr}={value}")
            for attr, value in self.__dict__.items()
        ]
        return f"{self.__class__.__name__}({', '.join(attributes)})"

    def __repr__(self) -> str:
        return self.__str__()

    @classmethod
    def _get_xml_tag(cls) -> str:
        return cls.__xml_name__ if getattr(cls, "__xml_name__", None) else cls.__name__

    @classmethod
    def from_xml_string(cls, string: str, convert_types=True, *, _warn_untracked=True):
        """ Creates an instance of the class from an XML text.

        Args:
            string (str): The XML text to parse.
            convert_types (bool, optional): Whether to convert the types of the fields. Defaults to True.
            _warn_untracked (bool, optional): Whether to warn if there are untracked children (for internal use only). Defaults to True.
        
        Raises:
            NotImplementedError: If the class does not have an `__xml_mapping__` attribute or if is not a dictionary.
            TypeError: If the value of the `__xml_mapping__` attribute is not valid.
            TypeConversionError: If the type of the field is not supported.
            
        Returns:
            XMLMapper: An instance of the class.
        """
        return cls.from_xml(
            ET.fromstring(string),
            convert_types=convert_types,
            _warn_untracked=_warn_untracked,
        )

    @classmethod
    def from_xml(cls, element: ET._Element, convert_types=True, *, _warn_untracked=True):
        """ Creates an instance of the class from an XML text.
        
        Args:
            element (ET._Element): The XML element to parse.
            convert_types (bool, optional): Whether to convert the types of the fields. Defaults to True.
            _warn_untracked (bool, optional): Whether to warn if there are untracked children (for internal use only). Defaults to True.
            
        Raises:
            NotImplementedError: If the class does not have an `__xml_mapping__` attribute or if is not a dictionary.
            TypeError: If the value of the `__xml_mapping__` attribute is not valid.
            TypeConversionError: If the type of the field is not supported.
            
        Returns:
            XMLMapper: An instance of the class.
        """
        if getattr(cls, "__xml_mapping__", None) is None and type(cls.__xml_mapping__) != dict:
            raise NotImplementedError(
                "This class does not have an __xml_mapping__ attribute defined."
            )
        
        # Check if there are tags that are not tracked in the `__xml_mapping__` attribute
        # I did not use a list comprehension since it would have been too long and unreadable
        child_tags = []
        for child in cls.__xml_mapping__.values():
            if type(child) == str:
                child_tags.append(child)
            elif isinstance(child, FieldGroup):
                child_tags.extend(list(child.target.__xml_mapping__.values()))
            elif isinstance(child, Field):
                if getattr(child, "is_parent"):
                    child_tags.append(child.tag)
                elif getattr(child, "tag", None):
                    child_tags.append(child.tag)
                else:
                    child_tags.append(child.target._get_xml_tag())

        untracked_children = [
            child.tag for child in element.iterchildren() if child.tag not in child_tags
        ]

        if untracked_children and _warn_untracked:
            print(
                f"\nWARNING: A total of {len(untracked_children)} children are not tracked ({', '.join(untracked_children)}) in the `{cls.__name__}.__xml_mapping__` class attribute.\n"
            )

        xml_object = cls()
        for attr, target in cls.__xml_mapping__.items():
            # Fail if the attribute is of the wrong type
            if type(target) != str and not isinstance(target, BaseField):
                raise TypeError(f"target must be a string or Field, not {type(target)}")

            if isinstance(target, FieldGroup):
                setattr(
                    xml_object,
                    attr,
                    target.target.from_xml(element, convert_types=convert_types, _warn_untracked=False),
                )

            elif isinstance(target, Field):
                child_class_name = ""
                if getattr(target, "tag", None):
                    child_class_name = target.tag
                elif getattr(target, "target", None):
                    child_class_name = target.target._get_xml_tag()

                if getattr(target, "is_parent"):
                    children = element.xpath(
                        f"{child_class_name}/{target.child.target._get_xml_tag()}"
                    )

                    children_obj = [
                        target.child.target.from_xml(child_xml, convert_types=convert_types, _warn_untracked=_warn_untracked)
                        for child_xml in children
                    ]

                    setattr(xml_object, attr, children_obj)
                else:
                    child_element = element.find(child_class_name)

                    if child_element is not None:
                        setattr(
                            xml_object,
                            attr,
                            target.target.from_xml(child_element, convert_types=convert_types, _warn_untracked=_warn_untracked),
                        )
            else:
                element_text = ""

                # ---> XML Attribute
                if target.strip().startswith("@"):
                    element_text = element.get(target[1:])

                elif target.strip().upper() == "#TEXT":
                    element_text = element.text

                # ---> XML Child Element
                else:
                    child_element = element.find(target)

                    if child_element is None:
                        continue

                    element_text = child_element.text

                if not convert_types:
                    setattr(xml_object, attr, element_text)
                    continue

                # =======> Type conversion <=======
                expected_type = get_type_hints(cls)[attr]

                try:
                    converted_value = None
                    if expected_type == bool or "[bool]" in str(expected_type):
                        if element_text is None:
                            element_text = ""
                        converted_value = element_text.lower() == "true"

                    elif expected_type == int or "[int]" in str(expected_type):
                        if element_text is None:
                            element_text = 0
                        converted_value = int(element_text)

                    elif expected_type == float or "[float]" in str(expected_type):
                        if element_text is None:
                            element_text = 0
                        converted_value = float(element_text)

                    elif expected_type == str or "[str]" in str(expected_type):
                        if element_text is None:
                            element_text = ""
                        converted_value = str(element_text)

                    else:
                        converted_value = element_text

                except ValueError:
                    raise TypeConversionError(
                        f"Error while converting `{cls.__name__}.{attr}`: `{element_text}` cannot be converted to `{expected_type.__name__}`."
                    )
                else:
                    setattr(xml_object, attr, converted_value)

        return xml_object

    # def to_xml(self) -> ET.Element:
    #     if getattr(self, "__xml_mapping__", None) is None or self.__xml_mapping__ == {}:
    #         raise NotImplementedError(
    #             "This class does not have an __xml_mapping__ attribute defined."
    #         )

    #     custom_element_name = getattr(self, "__xml_name__", "")
    #     root = ET.Element(
    #         custom_element_name if custom_element_name else self.__class__.__name__
    #     )

    #     for attr, xml_tag in self.__xml_mapping__.items():
    #         if isinstance(xml_tag, Field):
    #             tag_name = ""
    #             if getattr(xml_tag, "tag", None):
    #                 tag_name = xml_tag.tag
    #             elif getattr(xml_tag, "target", None):
    #                 if getattr(xml_tag.target, "__xml_name__", None):
    #                     tag_name = xml_tag.target.__xml_name__
    #                 else:
    #                     tag_name = xml_tag.target.__name__
    #             print(tag_name)

    #         if not hasattr(self, attr):
    #             raise AttributeError(
    #                 f"Attribute {attr} not found in class {self.__class__.__name__}"
    #             )

    #         attribute_value = getattr(self, attr)

    #         if attribute_value is None:
    #             continue

    #         # ---> XML Attribute
    #         if xml_tag.startswith("@"):
    #             root.set(xml_tag[1:], python2xml(attribute_value))

    #         # ---> XML Child Element
    #         else:
    #             child_element = ET.Element(xml_tag)
    #             if isinstance(attribute_value, list):
    #                 for item in attribute_value:
    #                     if isinstance(item, XMLMapper):
    #                         child_element.append(item.to_xml())
    #                     else:
    #                         raise TypeError(
    #                             f"Item {item} is not an instance of XMLConverter"
    #                         )

    #                 child_element[:] = sorted(
    #                     child_element, key=lambda child: child.tag
    #                 )
    #             else:
    #                 child_element.text = python2xml(attribute_value)

    #             root.append(child_element)

    #     root[:] = sorted(root, key=lambda child: child.tag)

    #     return root

    # def to_xml_string(
    #     self, short_empty_elements=False, prettify=False, indent="    "
    # ) -> str:
    #     """Converts the object to an XML string.

    #     Args:
    #         short_empty_elements (bool, optional): Shorten empty elements. Defaults to False.
    #         prettify (bool, optional): Pretty print the XML. Defaults to False.
    #         indent (str, optional): Character used while pretty printing the XML. Defaults to "    ".

    #     Returns:
    #         str: string representation of the XML object.
    #     """
    #     xml_string = ET.tostring(
    #         self.to_xml(),
    #         method="xml",
    #         encoding="UTF-8",
    #         short_empty_elements=short_empty_elements,
    #         xml_declaration=True,
    #     ).decode("utf-8")

    #     if not prettify:
    #         return xml_string

    #     pretty_xml = minidom.parseString(xml_string).toprettyxml(indent=indent)
    #     pretty_xml = _os.linesep.join([s for s in pretty_xml.splitlines() if s.strip()])

    #     return pretty_xml



import dataclasses
from typing import Optional

from easyfatt_db_connector.xml.company import Company
from easyfatt_db_connector.xml.document import Document
from easyfatt_db_connector.xml.common import Field, XMLMapper


@dataclasses.dataclass(init=False, repr=False)
class EasyfattXML(XMLMapper):
    """ Python object representation of the `.DefXml` file. 
    
	Each XML text gets encoded to its relative Python type (e.g. `false` becomes `False`).
    """
    
    __xml_name__ = "EasyfattDocuments"
    __xml_mapping__ = {
        "protocol_version": "@AppVersion",
        "creator_name": "@Creator",
        "creator_url": "@CreatorUrl",
        "company": Field(target=Company),
        "documents": Field(tag="Documents", is_parent=True, child=Field(Document)),
    }

    protocol_version: str = ""
    """ Versione del protocollo usato; va usato nell'integrazione e-commerce. 
    
    Example:
        ```xml
        <EasyfattDocuments AppVersion="[...]">...</EasyfattDocuments>
        ```
    """

    creator_name: str = ""
    """ Informazioni sul software o sul servizio che ha originato il file (nome della software house e/o del programma). 
    
    Example:
        ```xml
        <EasyfattDocuments Creator="[...]">...</EasyfattDocuments>
        ```
    """

    creator_url: str = ""
    """ Eventuale indirizzo web di approfondimento sul software o sul servizio che ha originato il file.
    
    Example:
        ```xml
        <EasyfattDocuments CreatorUrl="[...]">...</EasyfattDocuments>
        ```
    """

    company: Optional[Company] = None
    """ Informazioni sull'azienda che ha originato il file. 
    
    Example:
        ```xml
        <EasyfattDocuments>
            <Company>...</Company>
        </EasyfattDocuments>
        ```
    """

    documents: list[Document] = dataclasses.field(default_factory=list)
    """ Lista dei documenti contenuti nell'ordine. 
    
    Example:
        ```xml
        <EasyfattDocuments>
            <Documents>
                <Document>...</Document>
            </Documents>
        </EasyfattDocuments>
        ```
    """

    # def __init__(
    #     self,
    #     protocol_version: Optional[str] = None,
    #     creator_name: Optional[str] = None,
    #     creator_url: Optional[str] = None,
    # ):
    #     self.protocol_version = protocol_version
    #     self.creator_name = creator_name
    #     self.creator_url = creator_url

    #     self.company = None
    #     self.documents = []

    def add_document(self, document: Document):
        self.documents.append(document)

    def remove_document(self, document: Document):
        self.documents.remove(document)

    def set_company(self, company: Company):
        self.company = company

    def remove_company(self):
        self.company = None




""" Generic class for EasyfattDB. """
from pathlib import Path
from typing import Optional, Union
import urllib.parse

import fdb

import easyfatt_db_connector.utils.downloader as firebird_downloader
from easyfatt_db_connector.constants import (
    DEFAULT_DATABASE_CHARSET,
    DEFAULT_DATABASE_PASSWORD,
    DEFAULT_DATABASE_USERNAME,
    DEFAULT_FIREBIRD_LOCATION,
)
from easyfatt_db_connector.core.exceptions import FirebirdClientError


class EasyfattDBGeneric(object):
    """ Generic class for EasyfattDB. """
    archive_path: Path
    firebird_path: Path

    db_username: str
    db_password: str
    db_charset: Optional[str]

    def __init__(
        self,
        archive_path: Union[str, Path],
        firebird_path: Path = DEFAULT_FIREBIRD_LOCATION,
        download_firebird: bool = False,
        username: str = DEFAULT_DATABASE_USERNAME,
        password: str = DEFAULT_DATABASE_PASSWORD,
        charset: Optional[str] = DEFAULT_DATABASE_CHARSET,
    ) -> None:
        """Initialize a new EasyfattDB object and connect to the DB.

        Args:
            archive_path (str | Path): The path to the database file (with `.eft` extension).
            firebird_path (Path, optional): Path to the `fbembed.dll` parent folder. Defaults to DEFAULT_FIREBIRD_LOCATION.
            download_firebird (bool, optional): Whether the Firebird embedded driver should be automatically downloaded if not found. DO NOT USE IN PRODUCTION. Defaults to False.
            username (str, optional): Username used to connect to the database. Defaults to DEFAULT_DATABASE_USERNAME.
            password (str, optional): Password used to connect to the database. Defaults to DEFAULT_DATABASE_PASSWORD.
            charset (str | None, optional): Charset to use with the connection. Defaults to DEFAULT_DATABASE_CHARSET.
        """
        self.archive_path = Path(archive_path).expanduser().resolve()
        self.firebird_path = Path(firebird_path).expanduser().resolve()
        
        if not self.archive_path.exists():
            raise FirebirdClientError(f"The path '{self.archive_path}' does not exist.")
        
        # Check if the charset is supported
        if charset is None:
            self.db_charset = None
        elif charset.upper() in fdb.charset_map.keys():
            self.db_charset = charset
        elif charset in fdb.charset_map.values():
            inverted_map = {v: k for k, v in fdb.charset_map.items()}
            self.db_charset = inverted_map[charset]
        else:
            supported_charsets = ", ".join([str(charset) for charset in fdb.charset_map.keys()])
            raise FirebirdClientError(
                f"Character set '{charset}' not valid. Use one of '{supported_charsets}'"
            )

        self.db_username = username
        self.db_password = password
        self.db_charset = charset

        # Check if Firebase Embedded is installed
        try:
            if not self.firebird_path.exists():
                raise FirebirdClientError(f"The path '{self.firebird_path}' does not exist.")

            elif not (self.firebird_path / "fbembed.dll").exists():
                raise FirebirdClientError(
                    f"The path '{self.firebird_path}' MUST contain the file 'fbembed.dll'."
                )
        except FirebirdClientError:
            if download_firebird:
                print("Downloading Firebird Embedded...")
                firebird_downloader.download("R2_5_9", firebird_path=self.firebird_path)
            else:
                raise

    def __repr__(self) -> str:
        return f"<EasyfattDBGeneric: {self.archive_path}>"
    
    def get_connection_string(self, username=None, charset=None, protocol="firebird") -> str:
        """Returns the URI to connect to the database.
        
        Args:
            username (str, optional): Username used to connect to the database. Defaults to None.
            charset (str, optional): Charset to use with the connection. Defaults to None.
            protocol (str, optional): Protocol to use. Defaults to "firebird".
        """
        parameters = urllib.parse.urlencode({
            "charset": self.db_charset if charset is None else charset,
            "fb_library_name": str(self.firebird_path / "fbembed.dll"),
            "user": self.db_username if username is None else username
        }, quote_via=urllib.parse.quote)
        
        # 'localhost' MUST NOT be used, otherwise the connection will fail with error:
        #
        #       sqlalchemy.exc.DatabaseError: (fdb.fbcore.DatabaseError) ('Error while connecting to database:\n- SQLCODE: -902\n- Your 
        #       user name and password are not defined. Ask your database administrator to set up a Firebird login.', -902, 335544472)  
        return f"{protocol}://{self.db_username}:{self.db_password}@/{self.archive_path}?{parameters}"
    

    def is_locked(self):
        """ Check if the database is locked.
        
        TODO: Decide wether to make this method static or move it to the global scope and then reference it.
        This way could be parametrized to accept a path.

        FIXME: this method may cause your program to hang briefly (few seconds) when exiting.
        This behaviour could be caused by a global object that gets instantiated by the 
        `fdb.connect` call and it takes some time to be destroyed by the GC at the end.

        Returns:
            bool: Whether the database is locked or not.
        """
        try:
            # Use `WIN1252` instead of `UTF8` to fix error "SQLCODE: -204 block size exceeds implementation restriction"
            # See https://stackoverflow.com/q/40170882/8965861
            fdb.connect(
                database=str(self.archive_path),
                user=self.db_username,
                password=self.db_password if self.db_password else None,
                charset=self.db_charset,
                fb_library_name=str(self.firebird_path / "fbembed.dll"),
            )
        except (fdb.DatabaseError, UnicodeDecodeError) as e:
            error_message = str(e)
            try:
                sqlcode = e.args[1]
            except:
                sqlcode = None
            
            if "codec can't decode byte" in error_message or "lock manager error" in error_message:
                return True
            elif sqlcode == -902: # File used by another process (Error while connecting to database:\n- SQLCODE: -902\n- I/O error during "CreateFile (open)" operation for file "C:\\USERS\\{...}.EFT"\n- Error while trying to open file)
                return True
            else:
                raise
        except Exception as e:
            raise
        else:
            return False


from dataclasses import dataclass, field
from typing import Literal, Optional

from easyfatt_db_connector.xml.vat_code import VatCode

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

from easyfatt_db_connector.xml.product import Product
from easyfatt_db_connector.xml.common import Field, FieldGroup, XMLMapper

DocumentType: TypeAlias = Literal[
    "A",
    "B",
    "C",
    "D",
    "E",
    "F",
    "G",
    "H",
    "I",
    "J",
    "L",
    "M",
    "N",
    "O",
    "P",
    "Q",
    "R",
    "S",
]


@dataclass(init=False, repr=False)
class Payment(XMLMapper):
    """Campi nelle righe di pagamento (elementi `<Payment>`)."""
    __xml_name__ = "Payment"

    __xml_mapping__ = {
        "advance": "Advance",
        "amount": "Amount",
        "date": "Date",
        "paid": "Paid",
    }

    advance: bool
    """ Se `true`, segnala che il pagamento è riferito ad un acconto.
    
    Example:
        ```xml
        <Payment>
            <Advance>false</Advance>
        </Payment>
        ```
    """

    amount: float
    """ Importo [Valuta].
    
    Example:
        ```xml
        <Payment>
            <Amount>100</Amount>
        </Payment>
        ```
    """

    date: str
    """ Data [Data].
    
    Example:
        ```xml
        <Payment>
            <Date>2010-01-31</Date>
        </Payment>
        ```
    """

    paid: bool
    """ Se "true", significa che il pagamento è stato eseguito.
    
    Example:
        ```xml
        <Payment>
            <Paid>false</Paid>
        </Payment>
        ```
    """
    
@dataclass(init=False, repr=False)
class DeliveryInfo(XMLMapper):
	""" Campi nelle righe di consegna (elementi `<Delivery*>`).
    
    I campi "Delivery" si riferiscono all'indirizzo di consegna o spedizione e
	vanno specificati se tale indirizzo differisce da quello nei precedenti campi "Customer".
	"""
    
	__xml_name__ = ""
        
	__xml_mapping__ = {
		"name": "DeliveryName",
		"address": "DeliveryAddress",
		"postcode": "DeliveryPostcode",
		"city": "DeliveryCity",
		"province": "DeliveryProvince",
		"country": "DeliveryCountry",
	}

	name: str = ""
	""" Nome e cognome o denominazione.
	
	Example:
		```xml
        <Document>
		    <DeliveryName>...</DeliveryName>
        </Document>
		```
	"""

	address: str = ""
	""" Indirizzo.
	
	Example:
		```xml
        <Document>
		    <DeliveryAddress>...</DeliveryAddress>
        </Document>
		```
	"""

	postcode: str = ""
	""" CAP.
	
	Example:
		```xml
        <Document>
		    <DeliveryPostcode>...</DeliveryPostcode>
        </Document>
		```
	"""

	city: str = ""
	""" Città.
	
	Example:
		```xml
        <Document>
		    <DeliveryCity>...</DeliveryCity>
        </Document>
		```
	"""

	province: str = ""
	""" Provincia (2 caratteri).
	
	Example:
		```xml
        <Document>
		    <DeliveryProvince>...</DeliveryProvince>
        </Document>
		```
	"""

	country: str = ""
	""" Nazione.
	
	Example:
		```xml
        <Document>
		    <DeliveryCountry>...</DeliveryCountry>
        </Document>
		```
	"""

@dataclass(init=False, repr=False)
class TransportInfo(XMLMapper):
    """ Campi nelle righe di trasporto (elementi `<Transport*>`, `<Shipment>`, ecc). """

    __xml_mapping__ = {
        "carrier": "Carrier",
        "reason": "TransportReason",
        "goods_appearance": "GoodsAppearance",
        "pieces": "NumOfPieces",
        "date_time": "TransportDateTime",
        "shipment_terms": "ShipmentTerms",
        "weight": "TransportedWeight",
        "tracking_number": "TrackingNumber",
    }

    carrier: str = ""
    """ Denominazione vettore.

    Example:
        ```xml
        <Document>
            <Carrier>...</Carrier>
        </Document>
        ```
    """

    reason: str = ""
    """ Causale trasporto.

    Example:
        ```xml
        <Document>
            <TransportReason>...</TransportReason>
        </Document>
        ```
    """

    goods_appearance: str = ""
    """ Aspetto delle merci.

    Example:
        ```xml
        <Document>
            <GoodsAppearance>...</GoodsAppearance>
        </Document>
        ```
    """

    pieces: str = "0"
    """ Numero colli.

    Example:
        ```xml
        <Document>
            <NumOfPieces>...</NumOfPieces>
        </Document>
        ```
    """

    date_time: str = ""
    """ Data e ora del trasporto.

    Example:
        ```xml
        <Document>
            <TransportDateTime>...</TransportDateTime>
        </Document>
        ```
    """

    shipment_terms: str = ""
    """ Porto (franco, assegnato...).

    Example:
        ```xml
        <Document>
            <ShipmentTerms>...</ShipmentTerms>
        </Document>
        ```
    """

    weight: str = "0"
    """ Peso trasportato.

    Example:
        ```xml
        <Document>
            <TransportedWeight>...</TransportedWeight>
        </Document>
        ```
    """

    tracking_number: str = ""
    """ Numero di tracciatura spedizione.

    Example:
        ```xml
        <Document>
            <TrackingNumber>...</TrackingNumber>
        </Document>
        ```
    """

@dataclass(init=False, repr=False)
class CustomerInfo(XMLMapper):
    """ Campi del'intestatario del documento. 
    
    Va notato che i campi "Customer" (ovvero ... "Cliente") conterranno invece i dati del fornitore per i documenti di fornitura (ordine cliente, arrivo merce, etc.).
    """

    __xml_mapping__ = {
        "code": "CustomerCode",
        "web_login": "CustomerWebLogin",
        "name": "CustomerName",
        "address": "CustomerAddress",
        "postcode": "CustomerPostcode",
        "city": "CustomerCity",
        "province": "CustomerProvince",
        "country": "CustomerCountry",
        "fiscal_code": "CustomerFiscalCode",
        "vat_code": "CustomerVatCode",
        "e_invoice_dest_code": "CustomerEInvoiceDestCode",
        "telephone": "CustomerTel",
        "mobile_phone": "CustomerCellPhone",
        "fax": "CustomerFax",
        "email": "CustomerEmail",
        "pec": "CustomerPec",
        "reference": "CustomerReference",
    }

    code: str = ""
    """ Codice anagrafica; serve per associare l'ordine all'eventuale cliente già inserito in anagrafica. 
    
    Durante l'importazione, in presenza di `<CustomerCode>` e in assenza altri codici `<Customer...>`, 
    i relativi campi anagrafica del documento saranno riempiti con i dati dell'anagrafica corrispondente 
    a `<CustomerCode>` e già codificata nell'archivio.

    Example:
        ```xml
        <Document>
            <CustomerCode>...</CustomerCode>
        </Document>
        ```
    """

    web_login: str = ""
    """ Login web del cliente (usato nell'integrazione e-commerce).
     
    Nella fase di identificazione dell'anagrafica, questo campo viene preso in considerazione solo in mancanza di `CustomerCode`.

    Example:
        ```xml
        <Document>
            <CustomerWebLogin>...</CustomerWebLogin>
        </Document>
        ```
    """

    name: str = ""
    """ Cognome e nome o denominazione sociale.

    Example:
        ```xml
        <Document>
            <CustomerName>...</CustomerName>
        </Document>
        ```
    """

    address: str = ""
    """ Indirizzo.

    Example:
        ```xml
        <Document>
            <CustomerAddress>...</CustomerAddress>
        </Document>
        ```
    """

    postcode: str = ""
    """ CAP.

    Example:
        ```xml
        <Document>
            <CustomerPostcode>...</CustomerPostcode>
        </Document>
        ```
    """

    city: str = ""
    """ Città.

    Example:
        ```xml
        <Document>
            <CustomerCity>...</CustomerCity>
        </Document>
        ```
    """

    province: str = ""
    """ Provincia (2 caratteri).

    Example:
        ```xml
        <Document>
            <CustomerProvince>...</CustomerProvince>
        </Document>
        ```
    """

    country: str = ""
    """ Nazione.

    Example:
        ```xml
        <Document>
            <CustomerCountry>...</CustomerCountry>
        </Document>
        ```
    """

    fiscal_code: str = ""
    """ Codice fiscale.

    Example:
        ```xml
        <Document>
            <CustomerFiscalCode>...</CustomerFiscalCode>
        </Document>
        ```
    """

    vat_code: str = ""
    """ Partita Iva.

    Example:
        ```xml
        <Document>
            <CustomerVatCode>...</CustomerVatCode>
        </Document>
        ```
    """

    e_invoice_dest_code: str = ""
    """ Codice destinatario o PEC per l'invio della fattura elettronica.

    Example:
        ```xml
        <Document>
            <CustomerEInvoiceDestCode>...</CustomerEInvoiceDestCode>
        </Document>
        ```
    """

    telephone: str = ""
    """ Numero di telefono.

    Example:
        ```xml
        <Document>
            <CustomerTel>...</CustomerTel>
        </Document>
        ```
    """

    mobile_phone: str = ""
    """ Numero di cellulare.

    Example:
        ```xml
        <Document>
            <CustomerCellPhone>...</CustomerCellPhone>
        </Document>
        ```
    """

    fax: str = ""
    """ Numero di fax.

    Example:
        ```xml
        <Document>
            <CustomerFax>...</CustomerFax>
        </Document>
        ```
    """

    email: str = ""
    """ Indirizzo e-mail.

    Example:
        ```xml
        <Document>
            <CustomerEmail>...</CustomerEmail>
        </Document>
        ```
    """

    pec: str = ""
    """ Indirizzo PEC.

    Example:
        ```xml
        <Document>
            <CustomerPec>...</CustomerPec>
        </Document>
        ```
    """
    
    reference: str = ""
    """ Persona di riferimento.

    Example:
        ```xml
        <Document>
            <CustomerReference>...</CustomerReference>
        </Document>
        ```
    """

@dataclass(init=False, repr=False)
class DocumentNotes(XMLMapper):
    """ Campi note/commenti del documento. """

    __xml_mapping__ = {
        "internal_comment": "InternalComment",
        "custom1": "CustomField1",
        "custom2": "CustomField2",
        "custom3": "CustomField3",
        "custom4": "CustomField4",
        "foot_note": "FootNotes",
    }

    internal_comment: str = ""
    """ Commento (nell'e-commerce, usare questo campo per indicare note libere dell'acquirente durante la fase dell'ordine).

    Example:
        ```xml
        <Document>
            <InternalComment>...</InternalComment>
        </Document>
        ```
    """

    custom1: str = ""
    """ Campo note 1.

    Example:
        ```xml
        <Document>
            <CustomField1>...</CustomField1>
        </Document>
        ```
    """

    custom2: str = ""
    """ Campo note 2.

    Example:
        ```xml
        <Document>
            <CustomField2>...</CustomField2>
        </Document>
        ```
    """

    custom3: str = ""
    """ Campo note 3.

    Example:
        ```xml
        <Document>
            <CustomField3>...</CustomField3>
        </Document>
        ```
    """

    custom4: str = ""
    """ Campo note 4.

    Example:
        ```xml
        <Document>
            <CustomField4>...</CustomField4>
        </Document>
        ```
    """

    foot_note: str = ""
    """ Note a fine pagina.

    Example:
        ```xml
        <Document>
            <FootNotes>...</CustomerCode>
        </Document>
        ```
    """

@dataclass(init=False, repr=False)
class WithholdingTax(XMLMapper):
    """ Campi ritenute d'acconto. """

    __xml_mapping__ = {
        "rate1": "WithholdingTaxPerc",
        "rate2": "WithholdingTaxPerc2",
        "amount": "WithholdingTaxAmount",
        "amount_extras": "WithholdingTaxAmountB",
        "name_extras": "WithholdingTaxNameB",
    }

    rate1: float = 0
    """ Percentuale ritenuta d'acconto applicata [Numerico].
    
    Example:
        ```xml
        <Document>
            <WithholdingTaxPerc>...</WithholdingTaxPerc>
        </Document>
        ```
    """

    rate2: float = 0
    """ Seconda percentuale della ritenuta d'acconto (es.: 20% del 50%) [Numerico].
    
    Example:
        ```xml
        <Document>
            <WithholdingTaxPerc2>...</WithholdingTaxPerc2>
        </Document>
        ```
    """

    amount: float = 0
    """ Totale ritenuta d'acconto calcolata nel documento [Valuta]. Ignorato in importazione.
    
    Example:
        ```xml
        <Document>
            <WithholdingTaxAmount>...</WithholdingTaxAmount>
        </Document>
        ```
    """

    amount_extras: float = 0
    """ Totale altre ritenute (es.: Enasarco) [Valuta]. Ignorato in importazione.
    
    Example:
        ```xml
        <Document>
            <WithholdingTaxAmountB>...</WithholdingTaxAmountB>
        </Document>
        ```
    """

    name_extras: str = ""
    """ Descrizione altre ritenute (es.: "Ritenuta ENASARCO").
    
    Example:
        ```xml
        <Document>
            <WithholdingTaxNameB>...</WithholdingTaxNameB>
        </Document>
        ```
    """


@dataclass(init=False, repr=False)
class Contributions(XMLMapper):
    """ Campi contributi previdenziali. """

    __xml_mapping__ = {
        "description": "ContribDescription",
        "percentage": "ContribPerc",
        "is_subject_to_withholding_tax": "ContribSubjectToWithholdingTax",
        "total": "ContribAmount",
        "vat_code": "ContribVatCode",
    }

    description: str = ""
    """ Descrizione contributi previdenziali.
    
    Example:
        ```xml
        <Document>
            <ContribDescription>...</ContribDescription>
        </Document>
        ```
    """
    
    percentage: float = 0
    """ Percentuale contributi previdenziali.
    
    Example:
        ```xml
        <Document>
            <ContribPerc>...</ContribPerc>
        </Document>
        ```
    """
    
    is_subject_to_withholding_tax: bool = False
    """ Contributi previdenziali soggetti a ritenuta d'acconto [true|false].
    
    Example:
        ```xml
        <Document>
            <ContribSubjectToWithholdingTax>false</ContribSubjectToWithholdingTax>
        </Document>
        ```
    """
    
    total: float = 0
    """ Ammontare contributi previdenziali. Ignorato in importazione.
    
    Example:
        ```xml
        <Document>
            <ContribAmount>...</ContribAmount>
        </Document>
        ```
    """
    
    vat_code: int = 0
    """ Aliquota Iva contributi previdenziali. Ignorato in importazione.
    
    Example:
        ```xml
        <Document>
            <ContribVatCode>...</ContribVatCode>
        </Document>
        ```
    """


@dataclass(init=False, repr=False)
class DocumentPDFFile(XMLMapper):
    """ Campo `<Pdf>`. """

    __xml_name__ = "Pdf"
    __xml_mapping__ = {
        "filename": "@FileName",
        "content": "#TEXT",
    }

    filename: str = ""
    """ Nome file PDF. 
    
    Example:
        ```xml
        <Document>
            <Pdf FileName="..."></Pdf>
        </Document>
        ```
    """

    content: str = ""
    """ Contenuto file codificato in base64.
    
    Example:
        ```xml
        <Document>
            <Pdf>...</Pdf>
        </Document>
        ```
    """
    

@dataclass(init=False, repr=False)
class Document(XMLMapper):
    """ Dati del documento corrente (elementi `<Document>`).

    Fonte: https://www.danea.it/software/easyfatt/xml/documenti/#section-124
    """

    __xml_mapping__ = {
        "date": "Date",
        "number": "Number",
        "rows": Field(tag="Rows", is_parent=True, child=Field(Product)),
        "payments": Field(tag="Payments", is_parent=True, child=Field(Payment)),
        "numbering": "Numbering",
        "type": "DocumentType",

        "delivery": FieldGroup(DeliveryInfo),
        "transport": FieldGroup(TransportInfo),
        "customer": FieldGroup(CustomerInfo),
        "notes": FieldGroup(DocumentNotes),

        "cost_description": "CostDescription",
        "cost_vat_code": Field(VatCode, tag="CostVatCode"),
        "cost_amount": "CostAmount",

        "total_without_tax": "TotalWithoutTax",
        "total_vat": "VatAmount",
        "total": "Total",
        "total_paid": "PaymentAdvanceAmount",
        
        "prices_include_vat": "PricesIncludeVat",
        "total_subject_to_withholding_tax": "TotalSubjectToWithholdingTax",

        "withholding_tax": FieldGroup(WithholdingTax),
        "contributions": FieldGroup(Contributions),

        # ----------------------------------

        "delayed_vat": "DelayedVat",
        "delayed_vat_description": "DelayedVatDesc",
        "delayed_vat_dueWithinOneYear": "DelayedVatDueWithinOneYear",


        "warehouse": "Warehouse",
        "price_list": "PriceList",
        "payment_name": "PaymentName",
        "payment_bank": "PaymentBank",
        "sales_agent": "SalesAgent",
        "expected_conclusion": "ExpectedConclusionDate",
        "reference": "DocReference",
        "pdf": Field(DocumentPDFFile, tag="Pdf"),
    }

    rows: list["Product"] = field(default_factory=list)
    """ Lista dei prodotti o delle note contenute nel documento. 
    
    Example:
        ```xml
        <Rows>
            <Row>...</Row>
        </Rows>
        ```
    """

    payments: list["Payment"] = field(default_factory=list)
    """ Lista di scadenze di pagamento del documento.
    
    Example:
        ```xml
        <Payments>
            <Payment>...</Payment>
        </Payments>
        ```
    """

    date: str = ""
    """ Data del documento. 
    
    Example:
        ```xml
        <Document>
            <Date>2010-03-23</Date>
        </Document>
        ```
    """

    number: str = ""
    """ Numero del documento. 
    
    Example:
        ```xml
        <Document>
            <Number>4</Number>
        </Document>
        ```
    """

    numbering: Optional[str] = ""
    """ Numerazione documento ( /a, /b, etc.). Se omesso, l'importazione userà la numerazione scelta nelle relative impostazioni.
    
    Example:
        ```xml
        <Document>
            <Numbering/>
        </Document>
        ```
    """

    type: Optional[DocumentType] = "C"
    """ Codice corrispondente al tipo di documento. Se `None`, il documento sarà considerato un ordine cliente. 
    
    Example:
        ```xml
        <Document>
            <DocumentType>C</DocumentType>
        </Document>
        ```
    """

    delivery: Optional[DeliveryInfo] = None
    """ Informazioni aggiuntive sulla consegna o spedizione.

    Example:
        ```xml
        <Document>
            <DeliveryName>...</DeliveryName>
            <DeliveryAddress>...</DeliveryAddress>
            <DeliveryPostcode>...</DeliveryPostcode>
            <DeliveryCity>...</DeliveryCity>
            <DeliveryProvince>...</DeliveryProvince>
            <DeliveryCountry>...</DeliveryCountry>
        </Document>
        ```
    """

    transport: Optional[TransportInfo] = None
    """ Informazioni aggiuntive sul trasporto.

    Example:
        ```xml
        <Document>
            <Carrier>...</Carrier>
            <TransportReason>...</TransportReason>
            <GoodsAppearance>...</GoodsAppearance>
            <NumOfPieces>...</NumOfPieces>
            <TransportDateTime>...</TransportDateTime>
            <ShipmentTerms>...</ShipmentTerms>
            <TransportedWeight>...</TransportedWeight>
            <TrackingNumber>...</TrackingNumber>
        </Document>
        ```
    """

    customer: Optional[CustomerInfo] = None
    """ Informazioni aggiuntive sull'intestatario del documento.
    
    Questo può essere cliente o fornitore a seconda del tipo di documento.

    Example:
        ```xml
        <Document>
            <CustomerCode>...</CustomerCode>
            <CustomerWebLogin>...</CustomerWebLogin>
            <CustomerName>...</CustomerName>
            <CustomerAddress>...</CustomerAddress>
            <CustomerVatCode>...</CustomerVatCode>
            <CustomerEmail>...</CustomerEmail>
        </Document>
        ```
    """

    notes: Optional[DocumentNotes] = None
    """ Note aggiuntive sul documento.

    Example:
        ```xml
        <Document>
            <InternalComment>...</InternalComment>
            <CustomField1>...</CustomField1>
            <CustomField2>...</CustomField2>
            <CustomField3>...</CustomField3>
            <CustomField4>...</CustomField4>
            <FootNotes>...</FootNotes>
        </Document>
        ```
    """

    cost_description: Optional[str] = None
    """ Descrizione spese aggiuntive (ad esempio per spese di trasporto).

    Example:
        ```xml
        <Document>
            <CostDescription>...</CostDescription>
        </Document>
        ```
    """

    cost_vat_code: Optional[VatCode] = None
    """ Codice IVA spese aggiuntive (deve essere già presente nella tabella "Categorie Iva" dell'applicazione).
    
    Il codice è accompagnato dalle seguenti proprietà, il cui uso è facoltativo:
    - `Perc` (percentuale di tassazione applicata)
    - `Class` (classe: imponibile, non imponibile, intra-ue, extra-ue, esente, escluso, fuori campo, iva non esposta, rev. charge)
    - `Description`: descrizione libera del codice Iva

    Example:
        ```xml
        <Document>
            <CostVatCode>...</CostVatCode>
        </Document>
        ```
    """

    cost_amount: Optional[float] = 0
    """ Importo spese aggiuntive.

    Example:
        ```xml
        <Document>
            <CostAmount>...</CostAmount>
        </Document>
        ```
    """

    total_without_tax: Optional[float] = 0
    """ Totale documento al netto dell'imposta IVA [Valuta]. Ignorato in fase di importazione.

    Example:
        ```xml
        <Document>
            <TotalWithoutTax>...</TotalWithoutTax>
        </Document>
        ```
    """

    total_vat: Optional[float] = 0
    """ Totale Iva calcolata [Valuta]. Ignorato in importazione.

    Example:
        ```xml
        <Document>
            <VatAmount>...</VatAmount>
        </Document>
        ```
    """

    total: float = 0
    """ Totale documento [Valuta]. Ignorato in importazione.

    Example:
        ```xml
        <Document>
            <Total>...</Total>
        </Document>
        ```
    """

    total_paid: float = 0
    """ Importo acconto già versato (va usato solo nei documenti che non prevedono l'indicazione dei singoli Payment).

    Example:
        ```xml
        <Document>
            <TotalPaid>...</TotalPaid>
        </Document>
        ```
    """

    prices_include_vat: Optional[bool] = None
    """ Se "true" i campi CostAmount e Row.Price vengono considerati importi ivati, altrimenti netti [true|false].

    Example:
        ```xml
        <Document>
            <PricesIncludeVat>true</PricesIncludeVat>
        </Document>
        ```
    """

    total_subject_to_withholding_tax: float = 0
    """ Totale imponibile per ritenuta d'acconto [Numerico].

    Example:
        ```xml
        <Document>
            <TotalSubjectToWithholdingTax>...</TotalSubjectToWithholdingTax>
        </Document>
        ```
    """

    withholding_tax: Optional[WithholdingTax] = None
    """ Informazioni sulla ritenuta d'acconto. """

    contributions: Optional[Contributions] = None
    """ Informazioni sui contributi previdenziali. """

    # ----------------------------------

    delayed_vat: Optional[bool] = None
    """ Iva ad esigibilità differita [true|false].

    Example:
        ```xml
        <Document>
            <DelayedVat>...</DelayedVat>
        </Document>
        ```
    """

    delayed_vat_description: str = ""
    """ Causale Iva ad esigibilità differita.

    Example:
        ```xml
        <Document>
            <DelayedVatDesc>...</DelayedVatDesc>
        </Document>
        ```
    """

    delayed_vat_dueWithinOneYear: Optional[bool] = None
    """ Iva ad esigibilità differita comunque dovuta dopo un anno [true|false].

    Example:
        ```xml
        <Document>
            <DelayedVatDueWithinOneYear>...</DelayedVatDueWithinOneYear>
        </Document>
        ```
    """


    warehouse: str = ""
    """ Denominazione del magazzino movimentato.

    Example:
        ```xml
        <Document>
            <Warehouse>...</Warehouse>
        </Document>
        ```
    """

    price_list: str = ""
    """ Denominazione del listino prezzi usato.

    Example:
        ```xml
        <Document>
            <PriceList>...</PriceList>
        </Document>
        ```
    """


    payment_name: str = ""
    """ Nome pagamento (deve essere già presente nella tabella "Tipi pagamento" di Easyfatt).

    Example:
        ```xml
        <Document>
            <PaymentName>...</PaymentName>
        </Document>
        ```
    """

    payment_bank: str = ""
    """ Banca pagamento.

    Example:
        ```xml
        <Document>
            <PaymentBank>...</PaymentBank>
        </Document>
        ```
    """

    sales_agent: str = ""
    """ Nome dell'agente venditore (deve corrispondere ad un nominativo già presente nella tabella degli agenti).

    Example:
        ```xml
        <Document>
            <SalesAgent>...</SalesAgent>
        </Document>
        ```
    """

    expected_conclusion: str = "2999-12-31"
    """ Data di conclusione prevista (usato esclusivamente negli ordini clienti e fornitori) [Data].

    Example:
        ```xml
        <Document>
            <ExpectedConclusionDate>...</ExpectedConclusionDate>
        </Document>
        ```
    """

    reference: str = ""
    """ Causale del documento.
    
    Example:
        ```xml
        <Document>
            <DocReference>...</DocReference>
        </Document>
        ```
    """

    pdf: Optional[DocumentPDFFile] = None
    """ Documento in formato Pdf condificato Base64.

    Example:
        ```xml
        <Document>
            <Pdf>...</Pdf>
        </Document>
        ```
    """

    # def __init__(
    #     self,
    #     date: str,
    #     number: str,
    #     rows: list[Product] = [],
    #     payments: list["Payment"] = [],
    #     numbering: Optional[str] = None,
    #     type: Optional[DocumentType] = None,
    # ):
    #     self.date = date
    #     self.number = number
    #     self.rows = rows
    #     self.payments = payments
    #     self.numbering = numbering
    #     self.type = type


if __name__ == "__main__":
    xml = """
        <Document>
            <CustomerCode>0058</CustomerCode>
            <CustomerWebLogin/>
            <CustomerName>Amici del Gioco</CustomerName>
            <CustomerAddress>Via Ctr. Di Suio, 56</CustomerAddress>
            <CustomerPostcode>04021</CustomerPostcode>
            <CustomerCity>Castelforte</CustomerCity>
            <CustomerProvince>LT</CustomerProvince>
            <CustomerCountry/>
            <CustomerVatCode>66982157196</CustomerVatCode>
            <CustomerTel>0771 656890</CustomerTel>
            <DocumentType>I</DocumentType>
            <Warehouse>Principale</Warehouse>
            <Date>2009-02-28</Date>
            <Number>72</Number>
            <Numbering/>
            <CostDescription/>
            <CostVatCode/>
            <CostAmount/>
            <TotalWithoutTax>511.6</TotalWithoutTax>
            <VatAmount>102.32</VatAmount>
            <WithholdingTaxAmount>0</WithholdingTaxAmount>
            <Total>613.92</Total>
            <PriceList>Privati</PriceList>
            <PricesIncludeVat>false</PricesIncludeVat>
            <WithholdingTaxPerc>0</WithholdingTaxPerc>
            <PaymentName>Bonifico 30-60 gg F.M.</PaymentName>
            <PaymentBank>Cassa di Risparmio - IBAN IT83 H062 2562 9610 7404 2366 76W</PaymentBank>
            <Payments>
                <Payment>
                    <Advance>false</Advance>
                    <Date>2009-03-31</Date>
                    <Amount>306.96</Amount>
                    <Paid>true</Paid>
                </Payment>
                <Payment>
                    <Advance>false</Advance>
                    <Date>2009-04-30</Date>
                    <Amount>306.96</Amount>
                    <Paid>true</Paid>
                </Payment>
            </Payments>
            <InternalComment/>
            <CustomField1/>
            <CustomField2/>
            <CustomField3/>
            <CustomField4/>
            <FootNotes/>
            <SalesAgent/>
            <Rows>
                <Row>
                    <Code>0011</Code>
                    <Description>Anello refrigerante</Description>
                    <Qty>5</Qty>
                    <Um>pz</Um>
                    <Price>5.24</Price>
                    <Discounts/>
                    <VatCode>20</VatCode>
                    <Total>26.2</Total>
                    <Stock>false</Stock>
                </Row>
                <Row>
                    <Code>0010</Code>
                    <Description>Bicchiere antigoccia</Description>
                    <Qty>5</Qty>
                    <Um>pz</Um>
                    <Price>4.26</Price>
                    <Discounts/>
                    <VatCode>20</VatCode>
                    <Total>21.3</Total>
                    <Stock>false</Stock>
                </Row>
                <Row>
                    <Code>0009</Code>
                    <Description>Cucchiaio silicone</Description>
                    <Qty>5</Qty>
                    <Um>pz</Um>
                    <Price>3.06</Price>
                    <Discounts/>
                    <VatCode>20</VatCode>
                    <Total>15.3</Total>
                    <Stock>false</Stock>
                </Row>
                <Row>
                    <Code>0008</Code>
                    <Description>Seggiolone</Description>
                    <Qty>2</Qty>
                    <Um>pz</Um>
                    <Price>130.56</Price>
                    <Discounts/>
                    <VatCode>20</VatCode>
                    <Total>261.12</Total>
                    <Stock>false</Stock>
                </Row>
                <Row>
                    <Code>0007</Code>
                    <Description>Seggiolino da tavolo</Description>
                    <Qty>2</Qty>
                    <Um>pz</Um>
                    <Price>66.64</Price>
                    <Discounts/>
                    <VatCode>20</VatCode>
                    <Total>133.28</Total>
                    <Stock>false</Stock>
                </Row>
                <Row>
                    <Code>0006</Code>
                    <Description>Posatine</Description>
                    <Qty>5</Qty>
                    <Um>pz</Um>
                    <Price>10.88</Price>
                    <Discounts/>
                    <VatCode>20</VatCode>
                    <Total>54.4</Total>
                    <Stock>false</Stock>
                </Row>
            </Rows>
        </Document>
    """
    
    document = Document.from_xml_string(xml)
    print(f"Example of <Document> as object: {document}\n")
    print(f"Document number is             : {document.number}\n")
    print(f"Document VatCode is            : {document.date}")
    print(f"Document VatCode percentage is : {document.type}")
    print(f"Document first product is      : {document.rows[0].description}")
    print(f"Document first payment is      : {document.payments[0].amount}\n")



from pathlib import Path


DEFAULT_FIREBIRD_LOCATION = Path("~/.cache/firebird-embedded/").expanduser()

DEFAULT_DATABASE_USERNAME = "SYSDBA"
DEFAULT_DATABASE_PASSWORD = "masterkey"
DEFAULT_DATABASE_CHARSET = "WIN1252"



from dataclasses import dataclass
from easyfatt_db_connector.xml.common import XMLMapper

@dataclass(init=False)
class Company(XMLMapper):
    """ Dati dell'azienda che ha originato il file. 
    
    Fonte: https://www.danea.it/software/easyfatt/xml/documenti/#section-126
    """
    __xml_mapping__ = {
        "name": "Name",
        "address": "Address",
        "postcode": "Postcode",
        "city": "City",
        "province": "Province",
        "country": "Country",
        "fiscal_code": "FiscalCode",
        "vat_code": "VatCode",
        "phone": "Tel",
        "fax": "Fax",
        "email": "Email",
        "homepage": "HomePage",
    }

    name: str = ""
    """ Nome del mittente (l'azienda che ha originato il file). 
    
    Example:
        ```xml
        <Company>
            <Name>Arredufficio Srl</Name>
        </Company>
    """

    address: str = ""
    """ Indirizzo dell'azienda.
    
    Example:
        ```xml
        <Company>
            <Address>Via Crissolo, 12</Address>
        </Company>
    """

    postcode: str = ""
    """ CAP dell'azienda.
    
    Example:
        ```xml
        <Company>
            <Postcode>10138</Postcode>
        </Company>
    """

    city: str = ""
    """ Città dell'azienda.
    
    Example:
        ```xml
        <Company>
            <City>Torino</City>
        </Company>
    """

    province: str = ""
    """ Provincia dell'azienda (2 caratteri).
    
    Example:
        ```xml
        <Company>
            <Province>TO</Province>
        </Company>
    """

    country: str = ""
    """ Nazione dell'azienda.
    
    Example:
        ```xml
        <Company>
            <Country>Italia</Country>
        </Company>
    """

    fiscal_code: str = ""
    """ Codice fiscale dell'azienda.
    
    Example:
        ```xml
        <Company>
            <FiscalCode>01303760282</FiscalCode>
        </Company>
    """

    vat_code: str = ""
    """ Partita IVA dell'azienda.
    
    Example:
        ```xml
        <Company>
            <VatCode>01303760282</VatCode>
        </Company>
    """

    phone: str = ""
    """ Numero di telefono dell'azienda.
    
    Example:
        ```xml
        <Company>
            <Tel>011-2568745</Tel>
        </Company>
    """

    fax: str = ""
    """ Numero di fax dell'azienda.
    
    Example:
        ```xml
        <Company>
            <Fax>011-2548793</Fax>
        </Company>
    """

    email: str = ""
    """ Indirizzo email dell'azienda.
    
    Example:
        ```xml
        <Company>
            <Email>arredufficiosrl@arredufficio.it</Email>
        </Company>
    """

    homepage: str = ""
    """ Sito web dell'azienda.
    
    Example:
        ```xml
        <Company>
            <HomePage>www.arredufficiosrl.it</HomePage>
        </Company>
    """


if __name__ == "__main__":
    xml = """
        <Company>
            <Name>Tuttobimbi Srl</Name>
            <Address>Via Armando Diaz, 162</Address>
            <Postcode>35010</Postcode>
            <City>Vigonza</City>
            <Province>PD</Province>
            <Country>Italia</Country>
            <FiscalCode>00165987261</FiscalCode>
            <VatCode>00165987261</VatCode>
            <Tel>049/1234567</Tel>
            <Fax>049/1234568</Fax>
            <Email>info@tutto-bimbi.com</Email>
            <HomePage>www.tutto-bimbi.com</HomePage>
        </Company>
    """
    
    company = Company.from_xml_string(xml)
    
    print(f"Example of <Company> as object: {company}\n")
    print(f"Company address is            : {company.address}\n")

    # TODO: fix this
    # print(f"Example of <Company> as XML   : {company.to_xml_string()}\n")


from dataclasses import dataclass


@dataclass
class BaseField(object):
    """ Basic implementation of an XML field.

    Allows to define relationships between XML elements and Python classes.
    """

    target: type = None
    """ Class to map the XML element to. """


@dataclass
class Field(BaseField):
    """ Represents an XML field.

    Allows to define relationships between XML elements and Python classes.
    """

    tag: str = None
    """ Override XML tag name. """

    is_parent: bool = False
    """ Whether the XML element is a list of elements. """

    child: "Field" = None
    """ Children XML elements. """


@dataclass
class FieldGroup(BaseField):
    """ This special field allows to group multiple XML fields together under the same dataclass.

    This is useful when many XML fields are related to each other and it is more convenient to
    assign them to the attributes of a single dataclass.
    """
    ...



from contextlib import contextmanager
import os
import shutil
import tempfile
from typing import Any, Generator, Literal, Optional
from pathlib import Path

import fdb

from easyfatt_db_connector.core.exceptions import DatabaseLockedError
from easyfatt_db_connector.core.connection._base import EasyfattDBGeneric

class TypedFDBConnection(fdb.Connection):
    """ Stub class used only to provide correct type hjinting to child classes,
    since the `fdb.Connection` class does not provide any type hinting for some return values.
    
    DO NOT USE THIS CLASS DIRECTLY: The methods are not implemented and will raise a NotImplementedError.
    """
    def execute_immediate(self) -> None:
        raise NotImplementedError

    def cursor(self) -> fdb.Cursor:
        raise NotImplementedError


class EasyfattFDB(EasyfattDBGeneric):
    def _connect(self, database_path: Path) -> TypedFDBConnection:
        """ Wrapper around `fdb.connect` which handles some edge cases.

        Args:
            database_path (Path): The path to the database file.

        Raises:
            FirebirdClientError: If the database is locked.
        """
        try:
            # Use `WIN1252` instead of `UTF8` to fix error "SQLCODE: -204 block size exceeds implementation restriction"
            # See https://stackoverflow.com/q/40170882/8965861
            return fdb.connect(
                database=str(database_path),
                user=self.db_username,
                password=self.db_password if self.db_password else None,
                charset=self.db_charset,
                fb_library_name=str(self.firebird_path / "fbembed.dll"),
            )
        except (fdb.DatabaseError, UnicodeDecodeError) as e:
            error_message = str(e)

            try:
                sqlcode = e.args[1]
            except:
                sqlcode = None
            
            if (
                ("codec can't decode byte" in error_message) or   # Not sure why, but sometimes this error pops up when the database is locked
                ("lock manager error" in error_message) or        # Database locked
                (sqlcode == -902)                                 # Database file used by another process
            ):
                raise DatabaseLockedError(
                    f"The database '{self.archive_path}' is locked. Close Easyfatt and try again."
                )
            
            # Error is not handled, propagate it to the caller
            raise
        
    """ Implementation of the `EasyfattDBGeneric` class using the `fdb` library. """
    @contextmanager
    def connect(self) -> Generator[TypedFDBConnection, None, None]:
        """Connect to the database and return a connection object.

        Can be used as a context manager.

        Yields:
            TypedFDBConnection: The connection object.
        
        Raises:
            FirebirdClientError: If the database is locked.
        """     
        handle, path = tempfile.mkstemp(prefix=f"{self.archive_path.stem}-", suffix=".eft.tmp~")
        os.close(handle)

        temp_database = Path(path)
        shutil.copy(self.archive_path, temp_database)

        connection = None
        try:
            connection: TypedFDBConnection = self._connect(temp_database)
            yield connection

        except Exception:
            raise

        finally:
            if connection is not None:
                connection.close()

            temp_database.unlink(missing_ok=True)



if __name__ == "__main__":
    database_path = Path("~/Documents/Danea Easyfatt/TestArchivio.eft").expanduser()

    db = EasyfattFDB(database_path)
    # print(f"- Is database locked: {db.is_locked()}")
    print(f"- Connection string : '{db.get_connection_string()}'")

    with db.connect() as connection:
        # Informazioni generali sul DB
        print(f"Version    : {connection.version}")
        print(f"Created on : {connection.creation_date}")
        print(f"DB Name    : {connection.database_name}")
        print(f"ReadOnly   : {connection.isreadonly()}")

        cur = connection.cursor()
        print(cur.execute("select * from RDB$ROLES").fetchall())
    
    print(f"The end.")


from __future__ import annotations
from contextlib import contextmanager
import time

from pathlib import Path
import shutil

from fdb.ibase import charset_map

import sqlalchemy
# import sqlalchemy_firebird

from easyfatt_db_connector.core.exceptions import FirebirdClientError
from easyfatt_db_connector.constants import (
    DEFAULT_DATABASE_CHARSET,
    DEFAULT_DATABASE_PASSWORD,
    DEFAULT_DATABASE_USERNAME,
    DEFAULT_FIREBIRD_LOCATION,
)


class EasyfattDB(object):
    """ Implementation of the `EasyfattDBGeneric` class using the `sqlalchemy` library. """
    archive_path: Path
    firebird_path: Path

    db_username: str
    db_password: str
    db_charset: str | None

    def __init__(
        self,
        archive_path: str | Path,
        firebird_path: Path = DEFAULT_FIREBIRD_LOCATION,
        db_user=DEFAULT_DATABASE_USERNAME,
        db_password=DEFAULT_DATABASE_PASSWORD,
        db_charset: str | None = DEFAULT_DATABASE_CHARSET,
    ):
        """Initialize a new EasyfattDB object and connect to the DB.

        Args:
            archive_path (str | Path): The path to the database
            firebird_path (Path, optional): The Firebird Embedded installation path
            db_user (str, optional): The user that will be used to connect to the database.
            db_password (str, optional): The password that will be used to connect to the database.
            db_charset (str | None, optional): The character set that will be used to interpret data.
        """
        self.archive_path = Path(archive_path).expanduser().resolve()
        self.firebird_path = Path(firebird_path).expanduser().resolve()
        
        if db_charset is None:
            self.db_charset = charset_map[None]

        elif db_charset.upper() in charset_map.keys():
            self.db_charset = charset_map[db_charset]
            
        elif db_charset in charset_map.values():
            self.db_charset = db_charset
        else:
            supported_charsets = ", ".join([str(charset) for charset in charset_map.keys()])

            raise FirebirdClientError(
                f"Character set '{db_charset}' not valid. Use one of '{supported_charsets}'"
            )

        self.db_username = db_user
        self.db_password = db_password

        # Check if Firebase Embedded is installed
        if not self.firebird_path.exists():
            raise FirebirdClientError(f"The path '{self.firebird_path}' does not exist.")

        if not (self.firebird_path / "fbembed.dll").exists():
            raise FirebirdClientError(
                f"The path '{self.firebird_path}' MUST contain the file 'fbembed.dll'."
            )


    def create_engine(self, username=None, password=None, database_path: Path=None):
        connection_url = sqlalchemy.URL.create(
            drivername="firebird",
            username=self.db_username if username is None else username,
            database=str(self.archive_path.resolve() if database_path is None else database_path.resolve()),
            query={
                "fb_library_name": str(self.firebird_path / "fbembed.dll"),
                "charset": self.db_charset,
                # "user": self.db_username if username is None else username
            }
        )
        return sqlalchemy.create_engine(connection_url, echo=False)
    
    @contextmanager
    def connect(self, engine: sqlalchemy.engine.base.Engine = None):
        """Connect to the database and return a connection object.

        Can be used as a context manager.

        Args:
            engine (sqlalchemy.engine.base.Engine, optional): The engine that will be used to connect to the database. Defaults to None.

        Yields:
            sqlalchemy.engine.base.Connection: The connection object.
        """
        temp_database = None

        if engine is None:
            temp_database = Path(f"{self.archive_path}.tmp~")
            shutil.copy(self.archive_path, temp_database)
    
            engine = self.create_engine(database_path=temp_database)

        try:
            with engine.connect() as connection:
                yield connection

            if temp_database is not None:
                engine.dispose()
            
        except sqlalchemy.exc.OperationalError as e:
            if "lock manager" in str(e):
                raise FirebirdClientError(
                    f"The database '{self.archive_path}' is locked. Close Easyfatt and try again."
                )
            else:
                raise
        finally:
            if temp_database is not None:
                try:
                    engine.dispose()
                except:
                    pass
                temp_database.unlink(missing_ok=True)



if __name__ == "__main__":
    print("Start")
    database_path = Path("~/Documents/Danea Easyfatt/TestArchivio.eft").expanduser()

    db = EasyfattDB(database_path)
    print(f"- Engine: {db.create_engine()}")
